# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Console for interacting with devices using HDLC.

To start the console, provide a serial port as the --device argument and paths
or globs for .proto files that define the RPC services to support:

  python -m pw_hdlc.rpc_console --device /dev/ttyUSB0 sample.proto

This starts an IPython console for communicating with the connected device. A
few variables are predefined in the interactive console. These include:

    rpcs   - used to invoke RPCs
    device - the serial device used for communication
    client - the pw_rpc.Client
    protos - protocol buffer messages indexed by proto package

An example echo RPC command:

  rpcs.pw.rpc.EchoService.Echo(msg="hello!")
"""

import argparse
import datetime
import glob
from inspect import cleandoc
import logging
from pathlib import Path
import sys
from types import ModuleType
from typing import (
    Any,
    Collection,
    Iterable,
    Iterator,
    List,
    Optional,
    Union,
)
import socket

import serial  # type: ignore

import pw_cli.log
import pw_console.python_logging
from pw_console import PwConsoleEmbed
from pw_console.pyserial_wrapper import SerialWithLogging
from pw_console.plugins.bandwidth_toolbar import BandwidthToolbar

from pw_log.proto import log_pb2
from pw_rpc.console_tools.console import flattened_rpc_completions
from pw_system.device import Device
from pw_tokenizer.database import LoadTokenDatabases
from pw_tokenizer.detokenize import Detokenizer
from pw_tokenizer import tokens

_LOG = logging.getLogger('tools')
_DEVICE_LOG = logging.getLogger('rpc_device')

PW_RPC_MAX_PACKET_SIZE = 256
SOCKET_SERVER = 'localhost'
SOCKET_PORT = 33000
MKFIFO_MODE = 0o666


def _parse_args():
    """Parses and returns the command line arguments."""
    parser = argparse.ArgumentParser(description=__doc__)
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-d', '--device', help='the serial port to use')
    parser.add_argument('-b',
                        '--baudrate',
                        type=int,
                        default=115200,
                        help='the baud rate to use')
    parser.add_argument(
        '--serial-debug',
        action='store_true',
        help=('Enable debug log tracing of all data passed through'
              'pyserial read and write.'))
    parser.add_argument(
        '-o',
        '--output',
        type=argparse.FileType('wb'),
        default=sys.stdout.buffer,
        help=('The file to which to write device output (HDLC channel 1); '
              'provide - or omit for stdout.'))
    parser.add_argument('--logfile', help='Console debug log file.')
    group.add_argument('-s',
                       '--socket-addr',
                       type=str,
                       help='use socket to connect to server, type default for\
            localhost:33000, or manually input the server address:port')
    parser.add_argument("--token-databases",
                        metavar='elf_or_token_database',
                        nargs="+",
                        action=LoadTokenDatabases,
                        help="Path to tokenizer database csv file(s).")
    parser.add_argument('--config-file',
                        type=Path,
                        help='Path to a pw_console yaml config file.')
    parser.add_argument('--proto-globs',
                        nargs='+',
                        help='glob pattern for .proto files')
    parser.add_argument('-v',
                        '--verbose',
                        action='store_true',
                        help='Enables debug logging when set')
    return parser.parse_args()


def _expand_globs(globs: Iterable[str]) -> Iterator[Path]:
    for pattern in globs:
        for file in glob.glob(pattern, recursive=True):
            yield Path(file)


def _start_ipython_terminal(device: Device,
                            serial_debug: bool = False,
                            config_file_path: Optional[Path] = None) -> None:
    """Starts an interactive IPython terminal with preset variables."""
    local_variables = dict(
        client=device.client,
        device=device,
        rpcs=device.rpcs,
        protos=device.client.protos.packages,
        # Include the active pane logger for creating logs in the repl.
        DEVICE_LOG=_DEVICE_LOG,
        LOG=logging.getLogger(),
    )

    welcome_message = cleandoc("""
        Welcome to the Pigweed Console!

        Help: Press F1 or click the [Help] menu
        To move focus: Press Shift-Tab or click on a window

        Example Python commands:

          device.rpcs.pw.rpc.EchoService.Echo(msg='hello!')
          LOG.warning('Message appears in Host Logs window.')
          DEVICE_LOG.warning('Message appears in Device Logs window.')
    """)

    client_info = device.info()
    completions = flattened_rpc_completions([client_info])

    log_windows = {
        'Device Logs': [_DEVICE_LOG],
        'Host Logs': [logging.getLogger()],
    }
    if serial_debug:
        log_windows['Serial Debug'] = [
            logging.getLogger('pw_console.serial_debug_logger')
        ]

    interactive_console = PwConsoleEmbed(
        global_vars=local_variables,
        local_vars=None,
        loggers=log_windows,
        repl_startup_message=welcome_message,
        help_text=__doc__,
        config_file_path=config_file_path,
    )
    interactive_console.hide_windows('Host Logs')
    interactive_console.add_sentence_completer(completions)
    if serial_debug:
        interactive_console.add_bottom_toolbar(BandwidthToolbar())

    # Setup Python logger propagation
    interactive_console.setup_python_logging()

    # Don't send device logs to the root logger.
    _DEVICE_LOG.propagate = False

    interactive_console.embed()


class SocketClientImpl:
    def __init__(self, config: str):
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        socket_server = ''
        socket_port = 0

        if config == 'default':
            socket_server = SOCKET_SERVER
            socket_port = SOCKET_PORT
        else:
            socket_server, socket_port_str = config.split(':')
            socket_port = int(socket_port_str)
        self.socket.connect((socket_server, socket_port))

    def write(self, data: bytes):
        self.socket.sendall(data)

    def read(self, num_bytes: int = PW_RPC_MAX_PACKET_SIZE):
        return self.socket.recv(num_bytes)


def console(device: str,
            baudrate: int,
            proto_globs: Collection[str],
            token_databases: Collection[tokens.Database],
            socket_addr: str,
            logfile: str,
            output: Any,
            serial_debug: bool = False,
            config_file: Optional[Path] = None,
            verbose: bool = False) -> int:
    """Starts an interactive RPC console for HDLC."""
    # argparse.FileType doesn't correctly handle '-' for binary files.
    if output is sys.stdout:
        output = sys.stdout.buffer

    if not logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        logfile = pw_console.python_logging.create_temp_log_file()

    log_level = logging.DEBUG if verbose else logging.INFO
    pw_cli.log.install(log_level, True, False, logfile)
    _DEVICE_LOG.setLevel(log_level)
    _LOG.setLevel(log_level)

    detokenizer = None
    if token_databases:
        detokenizer = Detokenizer(tokens.Database.merged(*token_databases),
                                  show_errors=True)

    if not proto_globs:
        proto_globs = ['**/*.proto']

    protos: List[Union[ModuleType, Path]] = list(_expand_globs(proto_globs))

    # Append compiled log.proto library to avoid include errors when manually
    # provided, and shadowing errors due to ordering when the default global
    # search path is used.
    protos.append(log_pb2)

    if not protos:
        _LOG.critical('No .proto files were found with %s',
                      ', '.join(proto_globs))
        _LOG.critical('At least one .proto file is required')
        return 1

    _LOG.debug('Found %d .proto files found with %s', len(protos),
               ', '.join(proto_globs))

    serial_impl = serial.Serial
    if serial_debug:
        serial_impl = SerialWithLogging

    timestamp_decoder = None
    if socket_addr is None:
        serial_device = serial_impl(device, baudrate, timeout=1)
        read = lambda: serial_device.read(8192)
        write = serial_device.write

        # Overwrite decoder for serial device.
        def milliseconds_to_string(timestamp):
            """Parses milliseconds since boot to a human-readable string."""
            return str(datetime.timedelta(seconds=timestamp / 1e3))[:-3]

        timestamp_decoder = milliseconds_to_string
    else:
        try:
            socket_device = SocketClientImpl(socket_addr)
            read = socket_device.read
            write = socket_device.write
        except ValueError:
            _LOG.exception('Failed to initialize socket at %s', socket_addr)
            return 1

    device_client = Device(1,
                           read,
                           write,
                           protos,
                           detokenizer,
                           timestamp_decoder=timestamp_decoder,
                           rpc_timeout_s=5)

    _start_ipython_terminal(device_client, serial_debug, config_file)
    return 0


def main() -> int:
    return console(**vars(_parse_args()))


if __name__ == '__main__':
    sys.exit(main())



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class HashFunction(_HashFunction, metaclass=_HashFunctionEnumTypeWrapper):
    pass
class _HashFunction:
    V = typing.NewType('V', builtins.int)
class _HashFunctionEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_HashFunction.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    # Never use this in any TUF metadata.
    UNKNOWN_HASH_FUNCTION = HashFunction.V(0)
    SHA256 = HashFunction.V(1)

# Never use this in any TUF metadata.
UNKNOWN_HASH_FUNCTION = HashFunction.V(0)
SHA256 = HashFunction.V(1)
global___HashFunction = HashFunction


class KeyType(_KeyType, metaclass=_KeyTypeEnumTypeWrapper):
    pass
class _KeyType:
    V = typing.NewType('V', builtins.int)
class _KeyTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_KeyType.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    # Never use this in any TUF metadata.
    UNKNOWN_KEY_TYPE = KeyType.V(0)
    RSA = KeyType.V(1)
    ED25519 = KeyType.V(2)
    ECDSA_SHA2_NISTP256 = KeyType.V(3)

# Never use this in any TUF metadata.
UNKNOWN_KEY_TYPE = KeyType.V(0)
RSA = KeyType.V(1)
ED25519 = KeyType.V(2)
ECDSA_SHA2_NISTP256 = KeyType.V(3)
global___KeyType = KeyType


class KeyScheme(_KeyScheme, metaclass=_KeySchemeEnumTypeWrapper):
    pass
class _KeyScheme:
    V = typing.NewType('V', builtins.int)
class _KeySchemeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_KeyScheme.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    # Never use this in any TUF metadata.
    UNKNOWN_KEY_SCHEME = KeyScheme.V(0)
    # RSA Probabilistic signature scheme with appendix.
    # The underlying hash function is SHA256.
    # In TUF Specification, this is referred to as "rsassa-pss-sha256".
    RSASSA_PSS_SHA256_SCHEME = KeyScheme.V(1)
    # Elliptic Curve digital signature algorithm based on Twisted Edwards curves.
    # See https://ed25519.cr.yp.to/.
    # In TUF Specification, it is referred to as "ed25519".
    ED25519_SCHEME = KeyScheme.V(2)
    # Elliptic Curve Digital Signature Algorithm with NIST P-256 curve signing
    # and SHA-256 hashing. See
    # https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm In
    # TUF Specification, it is referred to as "ecdsa-sha2-nistp256".
    ECDSA_SHA2_NISTP256_SCHEME = KeyScheme.V(3)

# Never use this in any TUF metadata.
UNKNOWN_KEY_SCHEME = KeyScheme.V(0)
# RSA Probabilistic signature scheme with appendix.
# The underlying hash function is SHA256.
# In TUF Specification, this is referred to as "rsassa-pss-sha256".
RSASSA_PSS_SHA256_SCHEME = KeyScheme.V(1)
# Elliptic Curve digital signature algorithm based on Twisted Edwards curves.
# See https://ed25519.cr.yp.to/.
# In TUF Specification, it is referred to as "ed25519".
ED25519_SCHEME = KeyScheme.V(2)
# Elliptic Curve Digital Signature Algorithm with NIST P-256 curve signing
# and SHA-256 hashing. See
# https://en.wikipedia.org/wiki/Elliptic_Curve_Digital_Signature_Algorithm In
# TUF Specification, it is referred to as "ecdsa-sha2-nistp256".
ECDSA_SHA2_NISTP256_SCHEME = KeyScheme.V(3)
global___KeyScheme = KeyScheme


# Metadata for a particular TUF role (e.g. targets metadata).
# Was TufMetadata
class SignedRootMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    SERIALIZED_ROOT_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    # Serialized RootMetadata message that is the data portion of the metadata.
    serialized_root_metadata: builtins.bytes = ...
    # Signature of the canonical form of the role's serialized metadata
    # (serialized_root_metadata).
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]: ...
    def __init__(self,
        *,
        serialized_root_metadata : builtins.bytes = ...,
        signatures : typing.Optional[typing.Iterable[global___Signature]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"serialized_root_metadata",b"serialized_root_metadata",u"signatures",b"signatures"]) -> None: ...
global___SignedRootMetadata = SignedRootMetadata

class SignedTimestampMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    SERIALIZED_TIMESTAMP_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    # Serialized TimestampMetadata message that is the data portion of the
    # metadata.
    serialized_timestamp_metadata: builtins.bytes = ...
    # Signature of the canonical form of the role's serialized metadata
    # (serialized_timestamp_metadata).
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]: ...
    def __init__(self,
        *,
        serialized_timestamp_metadata : builtins.bytes = ...,
        signatures : typing.Optional[typing.Iterable[global___Signature]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"serialized_timestamp_metadata",b"serialized_timestamp_metadata",u"signatures",b"signatures"]) -> None: ...
global___SignedTimestampMetadata = SignedTimestampMetadata

class SignedSnapshotMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    SERIALIZED_SNAPSHOT_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    # Serialized SnapshotMetadata message that is the data portion of the
    # metadata.
    serialized_snapshot_metadata: builtins.bytes = ...
    # Signature of the canonical form of the role's serialized metadata
    # (serialized_snapshot_metadata).
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]: ...
    def __init__(self,
        *,
        serialized_snapshot_metadata : builtins.bytes = ...,
        signatures : typing.Optional[typing.Iterable[global___Signature]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"serialized_snapshot_metadata",b"serialized_snapshot_metadata",u"signatures",b"signatures"]) -> None: ...
global___SignedSnapshotMetadata = SignedSnapshotMetadata

class SignedTargetsMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    SERIALIZED_TARGETS_METADATA_FIELD_NUMBER: builtins.int
    SIGNATURES_FIELD_NUMBER: builtins.int
    # Serialized TargetsMetadata message that is the data portion of the
    # metadata.
    serialized_targets_metadata: builtins.bytes = ...
    # Signature of the canonical form of the role's serialized metadata
    # (serialized_targets_metadata).
    @property
    def signatures(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Signature]: ...
    def __init__(self,
        *,
        serialized_targets_metadata : builtins.bytes = ...,
        signatures : typing.Optional[typing.Iterable[global___Signature]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"serialized_targets_metadata",b"serialized_targets_metadata",u"signatures",b"signatures"]) -> None: ...
global___SignedTargetsMetadata = SignedTargetsMetadata

class CommonMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    SPEC_VERSION_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    EXPIRES_FIELD_NUMBER: builtins.int
    ROLE_FIELD_NUMBER: builtins.int
    # Version number of the TUF Specification.
    # Follows the Semantic Versioning 2.0.0 (semver) format. Metadata is
    # written according to this version, and clients MUST verify that
    # "spec_version" matches the expected version number.
    # E.g. "1.0.0".
    spec_version: typing.Text = ...
    # Metadata file version.
    # Clients MUST NOT replace a metadata file with a version number less than
    # the one currently trusted.
    version: builtins.int = ...
    # Expiration time for the metadata.
    # Indicates when this metadata should be considered expired and no longer
    # trusted by clients. Notice the TUF Specification defines this as a JSON
    # string following the ISO 8601 standard. The expected format of the date and
    # time string is "YYYY-MM-DDTHH:MM:SSZ". Time is always in UTC, and the "Z"
    # time zone designator is attached to indicate a zero UTC offset.
    # E.g. "2030-08-26T16:48:27Z".
    @property
    def expires(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    # Role type for the metadata.
    # Indicates the type of the metadata. Valid values are 'root', 'targets',
    # 'snapshot' and 'timestamp' as defined in the TUF spec, though we don't
    # plan to support 'mirrors'.
    #
    # This field serves as a "magic code" that identifies a particular type of
    # a metadata. During verification, the client is expected to check this
    # field against the expected role type immediately after verifying the
    # signatures of a metadata. This can be considered a "confidence booster"
    # in the absence of canonical protobuf -- i.e. it makes the various
    # `serialized_x_metadata` fields more tamper resistant.
    role: typing.Text = ...
    def __init__(self,
        *,
        spec_version : typing.Text = ...,
        version : builtins.int = ...,
        expires : typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        role : typing.Text = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"_expires",b"_expires",u"_role",b"_role",u"expires",b"expires",u"role",b"role"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"_expires",b"_expires",u"_role",b"_role",u"expires",b"expires",u"role",b"role",u"spec_version",b"spec_version",u"version",b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_expires",b"_expires"]) -> typing.Optional[typing_extensions.Literal["expires"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_role",b"_role"]) -> typing.Optional[typing_extensions.Literal["role"]]: ...
global___CommonMetadata = CommonMetadata

# This content is signed.
class RootMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_METADATA_FIELD_NUMBER: builtins.int
    CONSISTENT_SNAPSHOT_FIELD_NUMBER: builtins.int
    KEYS_FIELD_NUMBER: builtins.int
    ROOT_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    TIMESTAMP_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    SNAPSHOT_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    TARGETS_SIGNATURE_REQUIREMENT_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    # Whether the repo supports consistent snapshots. If the repo has frequent
    # updates, you should set this to true.
    consistent_snapshot: builtins.bool = ...
    # Map from Keyid to Key.
    # Keyid is a unique identifier that identifies a cryptographic key.
    # Contains all of cryptographic keys used by this repository.
    @property
    def keys(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___KeyMapping]: ...
    # KeyConfig is the list of keys use for a particular role and the threshold.
    # Threshold is number of keys of that role whose signatures are required in
    # order to consider a file as being properly signed by that role.
    @property
    def root_signature_requirement(self) -> global___SignatureRequirement: ...
    @property
    def timestamp_signature_requirement(self) -> global___SignatureRequirement: ...
    @property
    def snapshot_signature_requirement(self) -> global___SignatureRequirement: ...
    @property
    def targets_signature_requirement(self) -> global___SignatureRequirement: ...
    def __init__(self,
        *,
        common_metadata : typing.Optional[global___CommonMetadata] = ...,
        consistent_snapshot : builtins.bool = ...,
        keys : typing.Optional[typing.Iterable[global___KeyMapping]] = ...,
        root_signature_requirement : typing.Optional[global___SignatureRequirement] = ...,
        timestamp_signature_requirement : typing.Optional[global___SignatureRequirement] = ...,
        snapshot_signature_requirement : typing.Optional[global___SignatureRequirement] = ...,
        targets_signature_requirement : typing.Optional[global___SignatureRequirement] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"root_signature_requirement",b"root_signature_requirement",u"snapshot_signature_requirement",b"snapshot_signature_requirement",u"targets_signature_requirement",b"targets_signature_requirement",u"timestamp_signature_requirement",b"timestamp_signature_requirement"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"consistent_snapshot",b"consistent_snapshot",u"keys",b"keys",u"root_signature_requirement",b"root_signature_requirement",u"snapshot_signature_requirement",b"snapshot_signature_requirement",u"targets_signature_requirement",b"targets_signature_requirement",u"timestamp_signature_requirement",b"timestamp_signature_requirement"]) -> None: ...
global___RootMetadata = RootMetadata

# The timestamp role is used for freshness check of the snapshot. Any
# project-specific update metadata should go in the top-level targets_metadata
# or with the TargetFile information
class TimestampMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_METADATA_FIELD_NUMBER: builtins.int
    SNAPSHOT_METADATA_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    # Only one snapshot_metadata is used per timestamp.
    @property
    def snapshot_metadata(self) -> global___MetadataFile: ...
    def __init__(self,
        *,
        common_metadata : typing.Optional[global___CommonMetadata] = ...,
        snapshot_metadata : typing.Optional[global___MetadataFile] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"snapshot_metadata",b"snapshot_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"snapshot_metadata",b"snapshot_metadata"]) -> None: ...
global___TimestampMetadata = TimestampMetadata

# The snapshot role is used to ensure that the collection of targets_metadata
# files is securely consistent (no target metadata mix and match). Any
# project-specific update metadata should go in the top-level targets_metadata
# or with the TargetFile information
class SnapshotMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_METADATA_FIELD_NUMBER: builtins.int
    TARGETS_METADATA_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    # Map from Target metadata file name to MetadataFile.
    # File name can be an arbitrary name or a full file name with relative path.
    # This map should contain an entry for the top level targets role and all
    # delegated roles.
    @property
    def targets_metadata(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___MetadataFile]: ...
    def __init__(self,
        *,
        common_metadata : typing.Optional[global___CommonMetadata] = ...,
        targets_metadata : typing.Optional[typing.Iterable[global___MetadataFile]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"targets_metadata",b"targets_metadata"]) -> None: ...
global___SnapshotMetadata = SnapshotMetadata

# The targets role describes the target files that comprise the software
# update. Targets metadata is organized in to a top-level targets metadata file
# and optional multiple deligated targets metadata files
#
# The top-level targets metatdata is the correct place to put any
# project-specific build version information, including build ID, hardware rev,
# etc.
class TargetsMetadata(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    COMMON_METADATA_FIELD_NUMBER: builtins.int
    TARGET_FILES_FIELD_NUMBER: builtins.int
    @property
    def common_metadata(self) -> global___CommonMetadata: ...
    # Collection of target file information
    # Target file name can be an arbitrary name or a path that describes where
    # the file lives relative to the base directory of the repository, e.g.
    # "path/to/amber_tools/0".
    @property
    def target_files(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___TargetFile]: ...
    def __init__(self,
        *,
        common_metadata : typing.Optional[global___CommonMetadata] = ...,
        target_files : typing.Optional[typing.Iterable[global___TargetFile]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"common_metadata",b"common_metadata",u"target_files",b"target_files"]) -> None: ...
global___TargetsMetadata = TargetsMetadata

class Signature(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    KEY_ID_FIELD_NUMBER: builtins.int
    SIG_FIELD_NUMBER: builtins.int
    # Identifier of the key, which is bytes of the SHA-256 hash of the
    # canonical form of the key.
    key_id: builtins.bytes = ...
    # The signature of the canonical form of the role's serialized metadata
    # (serialized_{root,timestamp,snapshot,targets}_metadata).
    sig: builtins.bytes = ...
    def __init__(self,
        *,
        key_id : builtins.bytes = ...,
        sig : builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"key_id",b"key_id",u"sig",b"sig"]) -> None: ...
global___Signature = Signature

class KeyMapping(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    KEY_ID_FIELD_NUMBER: builtins.int
    KEY_FIELD_NUMBER: builtins.int
    # Identifier of the key, which is bytes of the SHA-256 hash of the
    # canonical form of the key.
    key_id: builtins.bytes = ...
    # Cryptographic key
    @property
    def key(self) -> global___Key: ...
    def __init__(self,
        *,
        key_id : builtins.bytes = ...,
        key : typing.Optional[global___Key] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"key",b"key"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"key",b"key",u"key_id",b"key_id"]) -> None: ...
global___KeyMapping = KeyMapping

# Identifies an asymmetric cryptographic key.
class Key(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    KEY_TYPE_FIELD_NUMBER: builtins.int
    SCHEME_FIELD_NUMBER: builtins.int
    KEYVAL_FIELD_NUMBER: builtins.int
    # Denotes a public key signature system, such as RSA or ECDSA.
    key_type: global___KeyType.V = ...
    # Denotes the signature scheme corresponding to the key type. For example:
    # "rsassa-pss-sha256" or "ecdsa-sha2-nistp256".
    scheme: global___KeyScheme.V = ...
    # Stores the serialized public key for this cryptographic algorithm.
    keyval: builtins.bytes = ...
    def __init__(self,
        *,
        key_type : global___KeyType.V = ...,
        scheme : global___KeyScheme.V = ...,
        keyval : builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"key_type",b"key_type",u"keyval",b"keyval",u"scheme",b"scheme"]) -> None: ...
global___Key = Key

# The set of cryptographic keys used by a specific role. For example, list of
# key_ids used by the top level role "root".
class SignatureRequirement(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    KEY_IDS_FIELD_NUMBER: builtins.int
    THRESHOLD_FIELD_NUMBER: builtins.int
    # Set of Keyid's.
    # Keyid is a unique identifier that identifies a cryptographic key.
    # E.g. "f2d5020d08aea06a0a9192eb6a4f549e17032ebefa1aa9ac167c1e3e727930d6".
    @property
    def key_ids(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    # Threshold of signatures required to trust given file.
    # In other words; the number of keys of that role whose signatures are
    # required in order to consider a file as being properly signed by that role.
    threshold: builtins.int = ...
    def __init__(self,
        *,
        key_ids : typing.Optional[typing.Iterable[builtins.bytes]] = ...,
        threshold : builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"key_ids",b"key_ids",u"threshold",b"threshold"]) -> None: ...
global___SignatureRequirement = SignatureRequirement

class Hash(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FUNCTION_FIELD_NUMBER: builtins.int
    HASH_FIELD_NUMBER: builtins.int
    function: global___HashFunction.V = ...
    # Digest of the cryptographic hash function computed on the target file.
    hash: builtins.bytes = ...
    def __init__(self,
        *,
        function : global___HashFunction.V = ...,
        hash : builtins.bytes = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"function",b"function",u"hash",b"hash"]) -> None: ...
global___Hash = Hash

# Descriptor for a file stored in this repository. Linked to from target
# metadata.
class TargetFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FILE_NAME_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    HASHES_FIELD_NUMBER: builtins.int
    # Target file name can be an arbitrary name or a path that describes where
    # the file lives relative to the base directory of the repository, e.g.
    # "path/to/amber_tools/0".
    file_name: typing.Text = ...
    # Size of the target file (element payload) in bytes. This the size as stored
    # in the bundle. The final applied size can be different due to optional
    # compression.
    length: builtins.int = ...
    # Map from algorithm name to Hash.
    # Algorithm name is the name of a cryptographic hash function. E.g. "sha256".
    # The Hash string is the hex digest of the cryptographic function computed on
    # the target file. E.g.
    # "65b8c67f51c993d898250f40aa57a317d854900b3a04895464313e48785440da".
    @property
    def hashes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Hash]: ...
    def __init__(self,
        *,
        file_name : typing.Text = ...,
        length : builtins.int = ...,
        hashes : typing.Optional[typing.Iterable[global___Hash]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"file_name",b"file_name",u"hashes",b"hashes",u"length",b"length"]) -> None: ...
global___TargetFile = TargetFile

class MetadataFile(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    FILE_NAME_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    LENGTH_FIELD_NUMBER: builtins.int
    HASHES_FIELD_NUMBER: builtins.int
    # Target file name can be an arbitrary name or a path that describes where
    # the file lives relative to the base directory of the repository, e.g.
    # "path/to/target/0".
    file_name: typing.Text = ...
    # Metadata file version. E.g. 3.
    version: builtins.int = ...
    # Size of the target file in bytes.
    length: builtins.int = ...
    # Map from algorithm name to Hash.
    # Algorithm name is the name of a cryptographic hash function. E.g. "sha256".
    # The Hash is the hex digest of the cryptographic function computed on the
    # target file. E.g.
    # "65b8c67f51c993d898250f40aa57a317d854900b3a04895464313e48785440da".
    @property
    def hashes(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Hash]: ...
    def __init__(self,
        *,
        file_name : typing.Text = ...,
        version : builtins.int = ...,
        length : builtins.int = ...,
        hashes : typing.Optional[typing.Iterable[global___Hash]] = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"_file_name",b"_file_name",u"_length",b"_length",u"file_name",b"file_name",u"length",b"length"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"_file_name",b"_file_name",u"_length",b"_length",u"file_name",b"file_name",u"hashes",b"hashes",u"length",b"length",u"version",b"version"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_file_name",b"_file_name"]) -> typing.Optional[typing_extensions.Literal["file_name"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_length",b"_length"]) -> typing.Optional[typing_extensions.Literal["length"]]: ...
global___MetadataFile = MetadataFile



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

class ThreadState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    class Enum(_Enum, metaclass=_EnumEnumTypeWrapper):
        pass
    class _Enum:
        V = typing.NewType('V', builtins.int)
    class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_Enum.V], builtins.type):
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
        # Thread state is invalid or cannot be expressed by this enum.
        UNKNOWN = ThreadState.Enum.V(0)
        # Interrupt handling is often done on a stack that isn't associated with a
        # true RTOS thread. This state indicates the provided thread info is for an
        # interrupt handler.
        INTERRUPT_HANDLER = ThreadState.Enum.V(1)
        # This is the currently active thread as marked by the RTOS. In crashes in
        # interrupt contexts, this isn’t necessarily the thread that crashed.
        RUNNING = ThreadState.Enum.V(2)
        # Thread is ready to run, but isn’t currently running.
        READY = ThreadState.Enum.V(3)
        # The thread is not ready to run, and will not be ready to run until it is
        # explicitly resumed.
        SUSPENDED = ThreadState.Enum.V(4)
        # The thread is waiting on something before it can run again.
        BLOCKED = ThreadState.Enum.V(5)
        # The thread is either not yet initialized, or has terminated. In other
        # words, this thread is a suspended thread that cannot be unsuspended.
        INACTIVE = ThreadState.Enum.V(6)

    # Thread state is invalid or cannot be expressed by this enum.
    UNKNOWN = ThreadState.Enum.V(0)
    # Interrupt handling is often done on a stack that isn't associated with a
    # true RTOS thread. This state indicates the provided thread info is for an
    # interrupt handler.
    INTERRUPT_HANDLER = ThreadState.Enum.V(1)
    # This is the currently active thread as marked by the RTOS. In crashes in
    # interrupt contexts, this isn’t necessarily the thread that crashed.
    RUNNING = ThreadState.Enum.V(2)
    # Thread is ready to run, but isn’t currently running.
    READY = ThreadState.Enum.V(3)
    # The thread is not ready to run, and will not be ready to run until it is
    # explicitly resumed.
    SUSPENDED = ThreadState.Enum.V(4)
    # The thread is waiting on something before it can run again.
    BLOCKED = ThreadState.Enum.V(5)
    # The thread is either not yet initialized, or has terminated. In other
    # words, this thread is a suspended thread that cannot be unsuspended.
    INACTIVE = ThreadState.Enum.V(6)

    def __init__(self,
        ) -> None: ...
global___ThreadState = ThreadState

class Thread(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    NAME_FIELD_NUMBER: builtins.int
    ACTIVE_FIELD_NUMBER: builtins.int
    STATE_FIELD_NUMBER: builtins.int
    RAW_BACKTRACE_FIELD_NUMBER: builtins.int
    SYMBOLIZED_BACKTRACE_FIELD_NUMBER: builtins.int
    RAW_STACK_FIELD_NUMBER: builtins.int
    STACK_START_POINTER_FIELD_NUMBER: builtins.int
    STACK_END_POINTER_FIELD_NUMBER: builtins.int
    STACK_POINTER_FIELD_NUMBER: builtins.int
    CPU_USAGE_HUNDREDTHS_FIELD_NUMBER: builtins.int
    STACK_POINTER_EST_PEAK_FIELD_NUMBER: builtins.int
    # Thread names must be unique; this allows extensions of Snapshot to augment
    # threads with additional data. This should either be human readable text, or
    # tokenized data (e.g. base-64 encoded or binary data).
    name: builtins.bytes = ...
    # This field has been deprecatdin favor of using the state enum to report
    # RUNNING or INTERRUPT_CONTEXT to mark them as active.
    #
    # Whether or not this thread is the thread is the currently active context
    # at the time of capture. For multi-thread dumps, this field should only be
    # set on ONE thread.
    active: builtins.bool = ...
    # A summarized thread state. RTOS-specific extensions of the Thread message
    # may provide more specific thread state information.
    state: global___ThreadState.Enum.V = ...
    # Contents of a stack trace. It is expected that this stack is pre-walked,
    # and contains addresses. Most recent stack events are at the beginning of
    # the captured stack trace.
    @property
    def raw_backtrace(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    # Results of symbolizing stack_entries. This is usually not provided by the
    # device, but instead by server/host side processing.
    @property
    def symbolized_backtrace(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[typing.Text]: ...
    # This should contain the raw contents of the thread's stack. This might not
    # match stack_size. It can be larger due to a stack overflow, or smaller due
    # to the implementation deciding to only capture a portion of the stack.
    # Partial stack captures are typically a result of storage/memory
    # limitations.
    raw_stack: builtins.bytes = ...
    # The address this thread's stack pointer began at. For descending stacks,
    # this is the highest address of the stack bounds. For ascending stacks, this
    # is the lowest address of the stack bounds.
    stack_start_pointer: builtins.int = ...
    # The furthest permitted address from where this thread's stack pointer
    # began. For descending stacks, this is the lowest address of the stack
    # bounds. For ascending stacks, this is the highest address of the stack
    # bounds.
    stack_end_pointer: builtins.int = ...
    # The current stack pointer of this thread.
    stack_pointer: builtins.int = ...
    # CPU usage info. This is the percentage of CPU time the thread has been
    # active in hundredths of a percent. (e.g. 5.00% = 500u)
    cpu_usage_hundredths: builtins.int = ...
    # The address of highest estimated currently used in the thread stack.
    # Percentage of bytes used can be calculated by:
    # (stack_estimate_max_addr-stack_start_pointer) /
    # (stack_end_pointer-stack_start_pointer) * 100%
    stack_pointer_est_peak: builtins.int = ...
    def __init__(self,
        *,
        name : builtins.bytes = ...,
        active : builtins.bool = ...,
        state : global___ThreadState.Enum.V = ...,
        raw_backtrace : typing.Optional[typing.Iterable[builtins.int]] = ...,
        symbolized_backtrace : typing.Optional[typing.Iterable[typing.Text]] = ...,
        raw_stack : builtins.bytes = ...,
        stack_start_pointer : builtins.int = ...,
        stack_end_pointer : builtins.int = ...,
        stack_pointer : builtins.int = ...,
        cpu_usage_hundredths : builtins.int = ...,
        stack_pointer_est_peak : builtins.int = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal[u"_cpu_usage_hundredths",b"_cpu_usage_hundredths",u"_stack_end_pointer",b"_stack_end_pointer",u"_stack_pointer",b"_stack_pointer",u"_stack_pointer_est_peak",b"_stack_pointer_est_peak",u"_stack_start_pointer",b"_stack_start_pointer",u"cpu_usage_hundredths",b"cpu_usage_hundredths",u"stack_end_pointer",b"stack_end_pointer",u"stack_pointer",b"stack_pointer",u"stack_pointer_est_peak",b"stack_pointer_est_peak",u"stack_start_pointer",b"stack_start_pointer"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"_cpu_usage_hundredths",b"_cpu_usage_hundredths",u"_stack_end_pointer",b"_stack_end_pointer",u"_stack_pointer",b"_stack_pointer",u"_stack_pointer_est_peak",b"_stack_pointer_est_peak",u"_stack_start_pointer",b"_stack_start_pointer",u"active",b"active",u"cpu_usage_hundredths",b"cpu_usage_hundredths",u"name",b"name",u"raw_backtrace",b"raw_backtrace",u"raw_stack",b"raw_stack",u"stack_end_pointer",b"stack_end_pointer",u"stack_pointer",b"stack_pointer",u"stack_pointer_est_peak",b"stack_pointer_est_peak",u"stack_start_pointer",b"stack_start_pointer",u"state",b"state",u"symbolized_backtrace",b"symbolized_backtrace"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_cpu_usage_hundredths",b"_cpu_usage_hundredths"]) -> typing.Optional[typing_extensions.Literal["cpu_usage_hundredths"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_stack_end_pointer",b"_stack_end_pointer"]) -> typing.Optional[typing_extensions.Literal["stack_end_pointer"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_stack_pointer",b"_stack_pointer"]) -> typing.Optional[typing_extensions.Literal["stack_pointer"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_stack_pointer_est_peak",b"_stack_pointer_est_peak"]) -> typing.Optional[typing_extensions.Literal["stack_pointer_est_peak"]]: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal[u"_stack_start_pointer",b"_stack_start_pointer"]) -> typing.Optional[typing_extensions.Literal["stack_start_pointer"]]: ...
global___Thread = Thread

# This message overlays the pw.snapshot.Snapshot proto. It's valid to encode
# this message to the same sink that a Snapshot proto is being written to.
class SnapshotThreadInfo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    THREADS_FIELD_NUMBER: builtins.int
    @property
    def threads(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Thread]: ...
    def __init__(self,
        *,
        threads : typing.Optional[typing.Iterable[global___Thread]] = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"threads",b"threads"]) -> None: ...
global___SnapshotThreadInfo = SnapshotThreadInfo



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console entry point."""

import argparse
import inspect
import logging
from pathlib import Path
import sys

import pw_cli.log
import pw_cli.argument_types

import pw_console
import pw_console.python_logging
from pw_console.log_store import LogStore
from pw_console.plugins.calc_pane import CalcPane
from pw_console.plugins.clock_pane import ClockPane

_LOG = logging.getLogger(__package__)
_ROOT_LOG = logging.getLogger('')


# TODO(tonymd): Remove this when no downstream projects are using it.
def create_temp_log_file():
    return pw_console.python_logging.create_temp_log_file()


def _build_argument_parser() -> argparse.ArgumentParser:
    """Setup argparse."""
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('-l',
                        '--loglevel',
                        type=pw_cli.argument_types.log_level,
                        default=logging.DEBUG,
                        help='Set the log level'
                        '(debug, info, warning, error, critical)')

    parser.add_argument('--logfile', help='Pigweed Console log file.')

    parser.add_argument('--test-mode',
                        action='store_true',
                        help='Enable fake log messages for testing purposes.')
    parser.add_argument('--config-file',
                        type=Path,
                        help='Path to a pw_console yaml config file.')
    parser.add_argument('--console-debug-log-file',
                        help='Log file to send console debug messages to.')

    return parser


def main() -> int:
    """Pigweed Console."""

    parser = _build_argument_parser()
    args = parser.parse_args()

    if not args.logfile:
        # Create a temp logfile to prevent logs from appearing over stdout. This
        # would corrupt the prompt toolkit UI.
        args.logfile = pw_console.python_logging.create_temp_log_file()

    pw_cli.log.install(level=args.loglevel,
                       use_color=True,
                       hide_timestamp=False,
                       log_file=args.logfile)

    if args.console_debug_log_file:
        pw_cli.log.install(level=logging.DEBUG,
                           use_color=True,
                           hide_timestamp=False,
                           log_file=args.console_debug_log_file,
                           logger=logging.getLogger('pw_console'))

    global_vars = None
    default_loggers = {}
    if args.test_mode:
        root_log_store = LogStore()
        _ROOT_LOG.addHandler(root_log_store)
        _ROOT_LOG.debug('pw_console test-mode starting...')

        fake_logger = logging.getLogger(
            pw_console.console_app.FAKE_DEVICE_LOGGER_NAME)
        default_loggers = {
            # Don't include pw_console package logs (_LOG) in the log pane UI.
            # Add the fake logger for test_mode.
            'Fake Device Logs': [fake_logger],
            'PwConsole Debug': [logging.getLogger('pw_console')],
            'All Logs': root_log_store,
        }
        # Give access to adding log messages from the repl via: `LOG.warning()`
        global_vars = dict(LOG=fake_logger)

    help_text = None
    app_title = None
    if args.test_mode:
        app_title = 'Console Test Mode'
        help_text = inspect.cleandoc("""
            Welcome to the Pigweed Console Test Mode!

            Example commands:

              rpcs.pw.rpc.EchoService.Echo(msg='hello!')

              LOG.warning('Message appears console log window.')
        """)

    console = pw_console.PwConsoleEmbed(
        global_vars=global_vars,
        loggers=default_loggers,
        test_mode=args.test_mode,
        help_text=help_text,
        app_title=app_title,
        config_file_path=args.config_file,
    )

    # Add example plugins used to validate behavior in the Pigweed Console
    # manual test procedure: https://pigweed.dev/pw_console/testing.html
    if args.test_mode:
        _ROOT_LOG.debug('pw_console.PwConsoleEmbed init complete')
        _ROOT_LOG.debug('Adding plugins...')
        console.add_window_plugin(ClockPane())
        console.add_window_plugin(CalcPane())
        _ROOT_LOG.debug('Starting prompt_toolkit full-screen application...')

    console.embed()

    if args.logfile:
        print(f'Logs saved to: {args.logfile}')

    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Classes for read and write transfers."""

import abc
import asyncio
from dataclasses import dataclass
import logging
import math
import threading
from typing import Any, Callable, Optional

from pw_status import Status
from pw_transfer.transfer_pb2 import Chunk

_LOG = logging.getLogger(__package__)


@dataclass(frozen=True)
class ProgressStats:
    bytes_sent: int
    bytes_confirmed_received: int
    total_size_bytes: Optional[int]

    def percent_received(self) -> float:
        if self.total_size_bytes is None:
            return math.nan

        return self.bytes_confirmed_received / self.total_size_bytes * 100

    def __str__(self) -> str:
        total = str(
            self.total_size_bytes) if self.total_size_bytes else 'unknown'
        return (f'{self.percent_received():5.1f}% ({self.bytes_sent} B sent, '
                f'{self.bytes_confirmed_received} B received of {total} B)')


ProgressCallback = Callable[[ProgressStats], Any]


class _Timer:
    """A timer which invokes a callback after a certain timeout."""
    def __init__(self, timeout_s: float, callback: Callable[[], Any]):
        self.timeout_s = timeout_s
        self._callback = callback
        self._task: Optional[asyncio.Task[Any]] = None

    def start(self, timeout_s: float = None) -> None:
        """Starts a new timer.

        If a timer is already running, it is stopped and a new timer started.
        This can be used to implement watchdog-like behavior, where a callback
        is invoked after some time without a kick.
        """
        self.stop()
        timeout_s = self.timeout_s if timeout_s is None else timeout_s
        self._task = asyncio.create_task(self._run(timeout_s))

    def stop(self) -> None:
        """Terminates a running timer."""
        if self._task is not None:
            self._task.cancel()
            self._task = None

    async def _run(self, timeout_s: float) -> None:
        await asyncio.sleep(timeout_s)
        self._task = None
        self._callback()


class Transfer(abc.ABC):
    """A client-side data transfer through a Manager.

    Subclasses are responsible for implementing all of the logic for their type
    of transfer, receiving messages from the server and sending the appropriate
    messages in response.
    """
    def __init__(self,
                 transfer_id: int,
                 send_chunk: Callable[[Chunk], None],
                 end_transfer: Callable[['Transfer'], None],
                 response_timeout_s: float,
                 initial_response_timeout_s: float,
                 max_retries: int,
                 progress_callback: ProgressCallback = None):
        self.id = transfer_id
        self.status = Status.OK
        self.done = threading.Event()

        self._send_chunk = send_chunk
        self._end_transfer = end_transfer

        self._retries = 0
        self._max_retries = max_retries
        self._response_timer = _Timer(response_timeout_s, self._on_timeout)
        self._initial_response_timeout_s = initial_response_timeout_s

        self._progress_callback = progress_callback

    async def begin(self) -> None:
        """Sends the initial chunk of the transfer."""
        self._send_chunk(self._initial_chunk())
        self._response_timer.start(self._initial_response_timeout_s)

    @property
    @abc.abstractmethod
    def data(self) -> bytes:
        """Returns the data read or written in this transfer."""

    @abc.abstractmethod
    def _initial_chunk(self) -> Chunk:
        """Returns the initial chunk to notify the sever of the transfer."""

    async def handle_chunk(self, chunk: Chunk) -> None:
        """Processes an incoming chunk from the server.

        Handles terminating chunks (i.e. those with a status) and forwards
        non-terminating chunks to handle_data_chunk.
        """
        self._response_timer.stop()
        self._retries = 0  # Received data from service, so reset the retries.

        _LOG.debug('Received chunk\n%s', str(chunk).rstrip())

        # Status chunks are only used to terminate a transfer. They do not
        # contain any data that requires processing.
        if chunk.HasField('status'):
            self.finish(Status(chunk.status))
            return

        await self._handle_data_chunk(chunk)

        # Start the timeout for the server to send a chunk in response.
        self._response_timer.start()

    @abc.abstractmethod
    async def _handle_data_chunk(self, chunk: Chunk) -> None:
        """Handles a chunk that contains or requests data."""

    @abc.abstractmethod
    def _retry_after_timeout(self) -> None:
        """Retries after a timeout occurs."""

    def _on_timeout(self) -> None:
        """Handles a timeout while waiting for a chunk."""
        if self.done.is_set():
            return

        self._retries += 1
        if self._retries > self._max_retries:
            self.finish(Status.DEADLINE_EXCEEDED)
            return

        _LOG.debug('Received no responses for %.3fs; retrying %d/%d',
                   self._response_timer.timeout_s, self._retries,
                   self._max_retries)
        self._retry_after_timeout()
        self._response_timer.start()

    def finish(self, status: Status, skip_callback: bool = False) -> None:
        """Ends the transfer with the specified status."""
        self._response_timer.stop()
        self.status = status

        if status.ok():
            total_size = len(self.data)
            self._update_progress(total_size, total_size, total_size)

        if not skip_callback:
            self._end_transfer(self)

        # Set done last so that the transfer has been fully cleaned up.
        self.done.set()

    def _update_progress(self, bytes_sent: int, bytes_confirmed_received: int,
                         total_size_bytes: Optional[int]) -> None:
        """Invokes the provided progress callback, if any, with the progress."""

        stats = ProgressStats(bytes_sent, bytes_confirmed_received,
                              total_size_bytes)
        _LOG.debug('Transfer %d progress: %s', self.id, stats)

        if self._progress_callback:
            self._progress_callback(stats)

    def _send_error(self, error: Status) -> None:
        """Sends an error chunk to the server and finishes the transfer."""
        self._send_chunk(Chunk(transfer_id=self.id, status=error.value))
        self.finish(error)


class WriteTransfer(Transfer):
    """A client -> server write transfer."""
    def __init__(
        self,
        transfer_id: int,
        data: bytes,
        send_chunk: Callable[[Chunk], None],
        end_transfer: Callable[[Transfer], None],
        response_timeout_s: float,
        initial_response_timeout_s: float,
        max_retries: int,
        progress_callback: ProgressCallback = None,
    ):
        super().__init__(transfer_id, send_chunk, end_transfer,
                         response_timeout_s, initial_response_timeout_s,
                         max_retries, progress_callback)
        self._data = data

        # Guard this class with a lock since a transfer parameters update might
        # arrive while responding to a prior update.
        self._lock = asyncio.Lock()
        self._offset = 0
        self._window_end_offset = 0
        self._max_chunk_size = 0
        self._chunk_delay_us: Optional[int] = None

        # The window ID increments for each parameters update.
        self._window_id = 0

        self._last_chunk = self._initial_chunk()

    @property
    def data(self) -> bytes:
        return self._data

    def _initial_chunk(self) -> Chunk:
        return Chunk(transfer_id=self.id)

    async def _handle_data_chunk(self, chunk: Chunk) -> None:
        """Processes an incoming chunk from the server.

        In a write transfer, the server only sends transfer parameter updates
        to the client. When a message is received, update local parameters and
        send data accordingly.
        """

        async with self._lock:
            self._window_id += 1
            window_id = self._window_id

            if not self._handle_parameters_update(chunk):
                return

            bytes_acknowledged = chunk.offset

        while True:
            if self._chunk_delay_us:
                await asyncio.sleep(self._chunk_delay_us / 1e6)

            async with self._lock:
                if self.done.is_set():
                    return

                if window_id != self._window_id:
                    _LOG.debug('Transfer %d: Skipping stale window', self.id)
                    return

                write_chunk = self._next_chunk()
                self._offset += len(write_chunk.data)
                sent_requested_bytes = self._offset == self._window_end_offset

            self._send_chunk(write_chunk)

            self._update_progress(self._offset, bytes_acknowledged,
                                  len(self.data))

            if sent_requested_bytes:
                break

        self._last_chunk = write_chunk

    def _handle_parameters_update(self, chunk: Chunk) -> bool:
        """Updates transfer state based on a transfer parameters update."""

        retransmit = True
        if chunk.HasField('type'):
            retransmit = chunk.type == Chunk.Type.PARAMETERS_RETRANSMIT

        if chunk.offset > len(self.data):
            # Bad offset; terminate the transfer.
            _LOG.error(
                'Transfer %d: server requested invalid offset %d (size %d)',
                self.id, chunk.offset, len(self.data))

            self._send_error(Status.OUT_OF_RANGE)
            return False

        if chunk.pending_bytes == 0:
            _LOG.error(
                'Transfer %d: service requested 0 bytes (invalid); aborting',
                self.id)
            self._send_error(Status.INTERNAL)
            return False

        if retransmit:
            # Check whether the client has sent a previous data offset, which
            # indicates that some chunks were lost in transmission.
            if chunk.offset < self._offset:
                _LOG.debug('Write transfer %d rolling back: offset %d from %d',
                           self.id, chunk.offset, self._offset)

            self._offset = chunk.offset

            # Retransmit is the default behavior for older versions of the
            # transfer protocol. The window_end_offset field is not guaranteed
            # to be set in these version, so it must be calculated.
            max_bytes_to_send = min(chunk.pending_bytes,
                                    len(self.data) - self._offset)
            self._window_end_offset = self._offset + max_bytes_to_send
        else:
            assert chunk.type == Chunk.Type.PARAMETERS_CONTINUE

            # Extend the window to the new end offset specified by the server.
            self._window_end_offset = min(chunk.window_end_offset,
                                          len(self.data))

        if chunk.HasField('max_chunk_size_bytes'):
            self._max_chunk_size = chunk.max_chunk_size_bytes

        if chunk.HasField('min_delay_microseconds'):
            self._chunk_delay_us = chunk.min_delay_microseconds

        return True

    def _retry_after_timeout(self) -> None:
        self._send_chunk(self._last_chunk)

    def _next_chunk(self) -> Chunk:
        """Returns the next Chunk message to send in the data transfer."""
        chunk = Chunk(transfer_id=self.id, offset=self._offset)
        max_bytes_in_chunk = min(self._max_chunk_size,
                                 self._window_end_offset - self._offset)

        chunk.data = self.data[self._offset:self._offset + max_bytes_in_chunk]

        # Mark the final chunk of the transfer.
        if len(self.data) - self._offset <= max_bytes_in_chunk:
            chunk.remaining_bytes = 0

        return chunk


class ReadTransfer(Transfer):
    """A client <- server read transfer.

    Although Python can effectively handle an unlimited transfer window, this
    client sets a conservative window and chunk size to avoid overloading the
    device. These are configurable in the constructor.
    """

    # The fractional position within a window at which a receive transfer should
    # extend its window size to minimize the amount of time the transmitter
    # spends blocked.
    #
    # For example, a divisor of 2 will extend the window when half of the
    # requested data has been received, a divisor of three will extend at a
    # third of the window, and so on.
    EXTEND_WINDOW_DIVISOR = 2

    def __init__(  # pylint: disable=too-many-arguments
            self,
            transfer_id: int,
            send_chunk: Callable[[Chunk], None],
            end_transfer: Callable[[Transfer], None],
            response_timeout_s: float,
            initial_response_timeout_s: float,
            max_retries: int,
            max_bytes_to_receive: int = 8192,
            max_chunk_size: int = 1024,
            chunk_delay_us: int = None,
            progress_callback: ProgressCallback = None):
        super().__init__(transfer_id, send_chunk, end_transfer,
                         response_timeout_s, initial_response_timeout_s,
                         max_retries, progress_callback)
        self._max_bytes_to_receive = max_bytes_to_receive
        self._max_chunk_size = max_chunk_size
        self._chunk_delay_us = chunk_delay_us

        self._remaining_transfer_size: Optional[int] = None
        self._data = bytearray()
        self._offset = 0
        self._pending_bytes = max_bytes_to_receive
        self._window_end_offset = max_bytes_to_receive

    @property
    def data(self) -> bytes:
        """Returns an immutable copy of the data that has been read."""
        return bytes(self._data)

    def _initial_chunk(self) -> Chunk:
        return self._transfer_parameters()

    async def _handle_data_chunk(self, chunk: Chunk) -> None:
        """Processes an incoming chunk from the server.

        In a read transfer, the client receives data chunks from the server.
        Once all pending data is received, the transfer parameters are updated.
        """

        if chunk.offset != self._offset:
            # Initially, the transfer service only supports in-order transfers.
            # If data is received out of order, request that the server
            # retransmit from the previous offset.
            self._send_chunk(self._transfer_parameters())
            return

        self._data += chunk.data
        self._pending_bytes -= len(chunk.data)
        self._offset += len(chunk.data)

        if chunk.HasField('remaining_bytes'):
            if chunk.remaining_bytes == 0:
                # No more data to read. Acknowledge receipt and finish.
                self._send_chunk(
                    Chunk(transfer_id=self.id, status=Status.OK.value))
                self.finish(Status.OK)
                return

            # The server may indicate if the amount of remaining data is known.
            self._remaining_transfer_size = chunk.remaining_bytes
        elif self._remaining_transfer_size is not None:
            # Update the remaining transfer size, if it is known.
            self._remaining_transfer_size -= len(chunk.data)

            # If the transfer size drops to zero, the estimate was inaccurate.
            if self._remaining_transfer_size <= 0:
                self._remaining_transfer_size = None

        total_size = None if self._remaining_transfer_size is None else (
            self._remaining_transfer_size + self._offset)
        self._update_progress(self._offset, self._offset, total_size)

        remaining_window_size = self._window_end_offset - self._offset
        extend_window = (remaining_window_size <= self._max_bytes_to_receive /
                         ReadTransfer.EXTEND_WINDOW_DIVISOR)

        if self._pending_bytes == 0:
            # All pending data was received. Send out a new parameters chunk for
            # the next block.
            self._send_chunk(self._transfer_parameters())
        elif extend_window:
            self._send_chunk(self._transfer_parameters(extend=True))

    def _retry_after_timeout(self) -> None:
        self._send_chunk(self._transfer_parameters())

    def _transfer_parameters(self, extend: bool = False) -> Chunk:
        """Sends an updated transfer parameters chunk to the server."""

        self._pending_bytes = self._max_bytes_to_receive
        self._window_end_offset = self._offset + self._max_bytes_to_receive

        chunk_type = (Chunk.Type.PARAMETERS_CONTINUE
                      if extend else Chunk.Type.PARAMETERS_RETRANSMIT)

        chunk = Chunk(transfer_id=self.id,
                      pending_bytes=self._pending_bytes,
                      window_end_offset=self._window_end_offset,
                      max_chunk_size_bytes=self._max_chunk_size,
                      offset=self._offset,
                      type=chunk_type)

        if self._chunk_delay_us:
            chunk.min_delay_microseconds = self._chunk_delay_us

        return chunk



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Module for running subprocesses from pw and capturing their output."""

import asyncio
import logging
import os
import shlex
import tempfile
from typing import IO, Tuple, Union, Optional

import pw_cli.color
import pw_cli.log

_COLOR = pw_cli.color.colors()
_LOG = logging.getLogger(__name__)

# Environment variable passed down to subprocesses to indicate that they are
# running as a subprocess. Can be imported by other code.
PW_SUBPROCESS_ENV = 'PW_SUBPROCESS'


class CompletedProcess:
    """Information about a process executed in run_async."""
    def __init__(self, process: 'asyncio.subprocess.Process',
                 output: Union[bytes, IO[bytes]]):
        assert process.returncode is not None
        self.returncode: int = process.returncode
        self.pid = process.pid
        self._output = output

    @property
    def output(self) -> bytes:
        # If the output is a file, read it, then close it.
        if not isinstance(self._output, bytes):
            with self._output as file:
                file.flush()
                file.seek(0)
                self._output = file.read()

        return self._output


async def _run_and_log(program: str, args: Tuple[str, ...], env: dict):
    process = await asyncio.create_subprocess_exec(
        program,
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
        env=env)

    output = bytearray()

    if process.stdout:
        while True:
            line = await process.stdout.readline()
            if not line:
                break

            output += line
            _LOG.log(pw_cli.log.LOGLEVEL_STDOUT, '[%s] %s',
                     _COLOR.bold_white(process.pid),
                     line.decode(errors='replace').rstrip())

    return process, bytes(output)


async def run_async(program: str,
                    *args: str,
                    log_output: bool = False,
                    timeout: Optional[float] = None) -> CompletedProcess:
    """Runs a command, capturing and optionally logging its output.

    Returns a CompletedProcess with details from the process.
    """

    _LOG.debug('Running `%s`',
               ' '.join(shlex.quote(arg) for arg in [program, *args]))

    env = os.environ.copy()
    env[PW_SUBPROCESS_ENV] = '1'
    output: Union[bytes, IO[bytes]]

    if log_output:
        process, output = await _run_and_log(program, args, env)
    else:
        output = tempfile.TemporaryFile()
        process = await asyncio.create_subprocess_exec(
            program,
            *args,
            stdout=output,
            stderr=asyncio.subprocess.STDOUT,
            env=env)

    try:
        await asyncio.wait_for(process.wait(), timeout)
    except asyncio.TimeoutError:
        _LOG.error('%s timed out after %d seconds', program, timeout)
        process.kill()
        await process.wait()

    if process.returncode:
        _LOG.error('%s exited with status %d', program, process.returncode)
    else:
        _LOG.info('%s exited successfully', program)

    return CompletedProcess(process, output)


def run(program: str, *args: str, **kwargs) -> CompletedProcess:
    """Synchronous wrapper for run_async."""
    return asyncio.run(run_async(program, *args, **kwargs))



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Decodes and detokenizes Base64-encoded strings in serial output.

The output is printed or saved to a file. Input is not supported.
"""

import argparse
import sys
from typing import BinaryIO, Iterable

import serial  # type: ignore
from pw_tokenizer import database, detokenize, tokens


def _parse_args():
    """Parses and return command line arguments."""

    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[database.token_databases_parser()])
    parser.add_argument('-d',
                        '--device',
                        required=True,
                        help='The serial device from which to read')
    parser.add_argument('-b',
                        '--baudrate',
                        type=int,
                        default=115200,
                        help='The baud rate for the serial device')
    parser.add_argument('-o',
                        '--output',
                        type=argparse.FileType('wb'),
                        default=sys.stdout.buffer,
                        help=('The file to which to write the output; '
                              'provide - or omit for stdout.'))
    parser.add_argument(
        '-p',
        '--prefix',
        default=detokenize.BASE64_PREFIX,
        help=('The one-character prefix that signals the start of a '
              'Base64-encoded message. (default: $)'))
    parser.add_argument(
        '-s',
        '--show_errors',
        action='store_true',
        help=('Show error messages instead of conversion specifiers when '
              'arguments cannot be decoded.'))

    return parser.parse_args()


def _detokenize_serial(databases: Iterable, device: serial.Serial,
                       baudrate: int, show_errors: bool, output: BinaryIO,
                       prefix: str) -> None:
    if output is sys.stdout:
        output = sys.stdout.buffer

    detokenizer = detokenize.Detokenizer(tokens.Database.merged(*databases),
                                         show_errors=show_errors)
    serial_device = serial.Serial(port=device, baudrate=baudrate)

    try:
        detokenizer.detokenize_base64_live(serial_device, output, prefix)
    except KeyboardInterrupt:
        output.flush()


def main():
    _detokenize_serial(**vars(_parse_args()))
    return 0


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates a setup.cfg file for a Python package."""

import argparse
from collections import defaultdict
import configparser
from dataclasses import dataclass
from itertools import chain
import json
from pathlib import Path
import sys
import textwrap
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Set,
    TextIO,
)

try:
    from pw_build.mirror_tree import mirror_paths
except ImportError:
    # Append this path to the module search path to allow running this module
    # before the pw_build package is installed.
    sys.path.append(str(Path(__file__).resolve().parent.parent))
    from pw_build.mirror_tree import mirror_paths


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--label', help='Label for this Python package')
    parser.add_argument('--proto-library',
                        dest='proto_libraries',
                        type=argparse.FileType('r'),
                        default=[],
                        action='append',
                        help='Paths')
    parser.add_argument('--generated-root',
                        required=True,
                        type=Path,
                        help='The base directory for the Python package')
    parser.add_argument('--setup-json',
                        required=True,
                        type=argparse.FileType('r'),
                        help='setup.py keywords as JSON')
    parser.add_argument('--module-as-package',
                        action='store_true',
                        help='Generate an __init__.py that imports everything')
    parser.add_argument('files',
                        type=Path,
                        nargs='+',
                        help='Relative paths to the files in the package')
    return parser.parse_args()


def _check_nested_protos(label: str, proto_info: Dict[str, Any]) -> None:
    """Checks that the proto library refers to this package."""
    python_package = proto_info['nested_in_python_package']

    if python_package != label:
        raise ValueError(
            f"{label}'s 'proto_library' is set to {proto_info['label']}, but "
            f"that target's 'python_package' is {python_package or 'not set'}. "
            f"Set {proto_info['label']}'s 'python_package' to {label}.")


@dataclass(frozen=True)
class _ProtoInfo:
    root: Path
    sources: Sequence[Path]
    deps: Sequence[str]


def _collect_all_files(
        root: Path, files: List[Path],
        paths_to_collect: Iterable[_ProtoInfo]) -> Dict[str, Set[str]]:
    """Collects files in output dir, adds to files; returns package_data."""
    root.mkdir(exist_ok=True)

    for proto_info in paths_to_collect:
        # Mirror the proto files to this package.
        files += mirror_paths(proto_info.root, proto_info.sources, root)

    # Find all subpackages, including empty ones.
    subpackages: Set[Path] = set()
    for file in (f.relative_to(root) for f in files):
        subpackages.update(root / path for path in file.parents)
    subpackages.remove(root)

    # Make sure there are __init__.py and py.typed files for each subpackage.
    for pkg in subpackages:
        for file in (pkg / name for name in ['__init__.py', 'py.typed']):
            if not file.exists():
                file.touch()
            files.append(file)

    pkg_data: Dict[str, Set[str]] = defaultdict(set)

    # Add all non-source files to package data.
    for file in (f for f in files if f.suffix != '.py'):
        pkg = file.parent

        package_name = pkg.relative_to(root).as_posix().replace('/', '.')
        pkg_data[package_name].add(file.name)

    return pkg_data


_PYPROJECT_FILE = '''\
# Generated file. Do not modify.
[build-system]
requires = ['setuptools', 'wheel']
build-backend = 'setuptools.build_meta'
'''


def _get_setup_keywords(pkg_data: dict, keywords: dict) -> Dict:
    """Gather all setuptools.setup() keyword args."""
    options_keywords = dict(
        packages=list(pkg_data),
        package_data={pkg: list(files)
                      for pkg, files in pkg_data.items()},
    )

    keywords['options'].update(options_keywords)
    return keywords


def _write_to_config(config: configparser.ConfigParser,
                     data: Dict,
                     section: Optional[str] = None):
    """Populate a ConfigParser instance with the contents of a dict."""
    # Add a specified section if missing.
    if section is not None and not config.has_section(section):
        config.add_section(section)

    for key, value in data.items():
        # Value is a dict so create a new subsection
        if isinstance(value, dict):
            _write_to_config(
                config,
                value,
                f'{section}.{key}' if section else key,
            )
        elif isinstance(value, list):
            if value:
                assert section is not None
                # Convert the list to an allowed str format.
                config[section][key] = '\n' + '\n'.join(value)
        else:
            assert section is not None
            # Add the value as a string. See expected types here:
            # https://setuptools.readthedocs.io/en/latest/userguide/declarative_config.html#specifying-values
            config[section][key] = str(value)


def _generate_setup_cfg(
    pkg_data: dict,
    keywords: dict,
    config_file_path: Path,
) -> None:
    """Creates a setup.cfg file based on setuptools keywords."""
    setup_keywords = _get_setup_keywords(pkg_data, keywords)

    config = configparser.ConfigParser()

    _write_to_config(config, setup_keywords)

    # Write the config to a file.
    with config_file_path.open('w') as config_file:
        config.write(config_file)


def _import_module_in_package_init(all_files: List[Path]) -> None:
    """Generates an __init__.py that imports the module.

    This makes an individual module usable as a package. This is used for proto
    modules.
    """
    sources = [
        f for f in all_files if f.suffix == '.py' and f.name != '__init__.py'
    ]
    assert len(sources) == 1, (
        'Module as package expects a single .py source file')

    source, = sources
    source.parent.joinpath('__init__.py').write_text(
        f'from {source.stem}.{source.stem} import *\n')


def _load_metadata(label: str,
                   proto_libraries: Iterable[TextIO]) -> Iterator[_ProtoInfo]:
    for proto_library_file in proto_libraries:
        info = json.load(proto_library_file)
        _check_nested_protos(label, info)

        deps = []
        for dep in info['dependencies']:
            with open(dep) as file:
                deps.append(json.load(file)['package'])

        yield _ProtoInfo(Path(info['root']),
                         tuple(Path(p) for p in info['protoc_outputs']), deps)


def main(generated_root: Path, files: List[Path], module_as_package: bool,
         setup_json: TextIO, label: str,
         proto_libraries: Iterable[TextIO]) -> int:
    """Generates a setup.py and other files for a Python package."""
    proto_infos = list(_load_metadata(label, proto_libraries))
    try:
        pkg_data = _collect_all_files(generated_root, files, proto_infos)
    except ValueError as error:
        msg = '\n'.join(textwrap.wrap(str(error), 78))
        print(
            f'ERROR: Failed to generate Python package {label}:\n\n'
            f'{textwrap.indent(msg, "  ")}\n',
            file=sys.stderr)
        return 1

    with setup_json:
        setup_keywords = json.load(setup_json)
        setup_keywords.setdefault('options', {})

    install_requires = setup_keywords['options'].setdefault(
        'install_requires', [])
    install_requires += chain.from_iterable(i.deps for i in proto_infos)

    if module_as_package:
        _import_module_in_package_init(files)

    # Create the pyproject.toml and setup.cfg files for this package.
    generated_root.joinpath('pyproject.toml').write_text(_PYPROJECT_FILE)
    _generate_setup_cfg(pkg_data,
                        setup_keywords,
                        config_file_path=generated_root.joinpath('setup.cfg'))

    return 0


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Build a Python Source tree."""

import argparse
import configparser
from datetime import datetime
import io
from pathlib import Path
import re
import shutil
import subprocess
import tempfile
from typing import Iterable

from pw_build.python_package import PythonPackage, load_packages


def _parse_args():
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('--tree-destination-dir',
                        type=Path,
                        help='Path to output directory.')
    parser.add_argument('--include-tests',
                        action='store_true',
                        help='Include tests in the tests dir.')

    parser.add_argument('--setupcfg-common-file',
                        type=Path,
                        help='A file containing the common set of options for'
                        'incluing in the merged setup.cfg provided version.')
    parser.add_argument('--setupcfg-version-append-git-sha',
                        action='store_true',
                        help='Append the current git SHA to the setup.cfg '
                        'version.')
    parser.add_argument('--setupcfg-version-append-date',
                        action='store_true',
                        help='Append the current date to the setup.cfg '
                        'version.')

    parser.add_argument(
        '--extra-files',
        nargs='+',
        help='Paths to extra files that should be included in the output dir.')

    parser.add_argument(
        '--input-list-files',
        nargs='+',
        type=Path,
        help='Paths to text files containing lists of Python package metadata '
        'json files.')

    return parser.parse_args()


class UnknownGitSha(Exception):
    "Exception thrown when the current git SHA cannot be found."


def get_current_git_sha() -> str:
    git_command = 'git log -1 --pretty=format:%h'
    process = subprocess.run(git_command.split(),
                             stdout=subprocess.PIPE,
                             stderr=subprocess.STDOUT)
    gitsha = process.stdout.decode()
    if process.returncode != 0 or not gitsha:
        error_output = f'\n"{git_command}" failed with:' f'\n{gitsha}'
        if process.stderr:
            error_output += f'\n{process.stderr.decode()}'
        raise UnknownGitSha('Could not determine the current git SHA.' +
                            error_output)
    return gitsha.strip()


def get_current_date() -> str:
    return datetime.now().strftime('%Y%m%d%H%M')


class UnexpectedConfigSection(Exception):
    "Exception thrown when the common configparser contains unexpected values."


def load_common_config(common_config: Path,
                       append_git_sha: bool = False,
                       append_date: bool = False) -> configparser.ConfigParser:
    """Load an existing ConfigParser file and update metadata.version."""
    config = configparser.ConfigParser()
    config.read(common_config)

    # Check for existing values that should not be present
    if config.has_option('options', 'packages'):
        value = str(config['options']['packages'])
        raise UnexpectedConfigSection(
            f'[options] packages already defined as: {value}')

    if config.has_section('options.package_data'):
        raise UnexpectedConfigSection(
            '[options.package_data] already defined as:\n' +
            str(dict(config['options.package_data'].items())))

    if config.has_section('options.entry_points'):
        raise UnexpectedConfigSection(
            '[options.entry_points] already defined as:\n' +
            str(dict(config['options.entry_points'].items())))

    # Metadata and option sections should already be defined.
    assert config.has_section('metadata')
    assert config.has_section('options')

    # Append build metadata if applicable.
    build_metadata = []
    if append_date:
        build_metadata.append(get_current_date())
    if append_git_sha:
        build_metadata.append(get_current_git_sha())
    if build_metadata:
        version_prefix = config['metadata']['version']
        build_metadata_text = '.'.join(build_metadata)
        config['metadata']['version'] = (
            f'{version_prefix}+{build_metadata_text}')
    return config


def update_config_with_packages(
    config: configparser.ConfigParser,
    python_packages: Iterable[PythonPackage],
) -> None:
    """Merge setup.cfg files from a set of python packages."""
    config['options']['packages'] = 'find:'
    config['options.package_data'] = {}
    config['options.entry_points'] = {}

    # Save a list of packages being bundled.
    included_packages = [pkg.package_name for pkg in python_packages]

    for pkg in python_packages:
        assert pkg.config

        # Collect install_requires
        if pkg.config.has_option('options', 'install_requires'):
            existing_requires = config['options'].get('install_requires', '\n')
            # Requires are delimited by newlines or semicolons.
            # Split existing list on either one.
            this_requires = re.split(r' *[\n;] *',
                                     pkg.config['options']['install_requires'])
            new_requires = existing_requires.splitlines() + this_requires
            # Remove requires already included in this merged config.
            new_requires = [
                line for line in new_requires
                if line and line not in included_packages
            ]
            # Remove duplictes and sort require list.
            new_requires_text = '\n' + '\n'.join(sorted(set(new_requires)))
            config['options']['install_requires'] = new_requires_text

        # Collect package_data
        if pkg.config.has_section('options.package_data'):
            for key, value in pkg.config['options.package_data'].items():
                config['options.package_data'][key] = value

        # Collect entry_points
        if pkg.config.has_section('options.entry_points'):
            for key, value in pkg.config['options.entry_points'].items():
                existing_entry_points = config['options.entry_points'].get(
                    key, '')
                new_entry_points = '\n'.join([existing_entry_points, value])
                # Remove any empty lines
                new_entry_points = new_entry_points.replace('\n\n', '\n')
                config['options.entry_points'][key] = new_entry_points


def write_config(
    common_config: Path,
    final_config: configparser.ConfigParser,
    tree_destination_dir: Path,
) -> None:
    """Write a the final setup.cfg file with license comment block."""
    # Get the license comment block from the common_config.
    comment_block_text = ''
    comment_block_match = re.search(r'((^#.*?[\r\n])*)([^#])',
                                    common_config.read_text(), re.MULTILINE)
    if comment_block_match:
        comment_block_text = comment_block_match.group(1)

    setup_cfg_file = tree_destination_dir.resolve() / 'setup.cfg'
    setup_cfg_text = io.StringIO()
    final_config.write(setup_cfg_text)
    setup_cfg_file.write_text(comment_block_text + setup_cfg_text.getvalue())


def build_python_tree(python_packages: Iterable[PythonPackage],
                      tree_destination_dir: Path,
                      include_tests: bool = False) -> None:
    """Install PythonPackages to a destination directory."""

    # Create the root destination directory.
    destination_path = tree_destination_dir.resolve()
    # Delete any existing files
    shutil.rmtree(destination_path, ignore_errors=True)
    destination_path.mkdir(exist_ok=True)

    # Define a temporary location to run setup.py build in.
    with tempfile.TemporaryDirectory() as build_base_name:
        build_base = Path(build_base_name)

        for pkg in python_packages:
            lib_dir_path = pkg.setuptools_build_with_base(
                build_base, include_tests=include_tests)

            # Move installed files from the temp build-base into
            # destination_path.
            for new_file in lib_dir_path.glob('*'):
                # Use str(Path) since shutil.move only accepts path-like objects
                # in Python 3.9 and up:
                #   https://docs.python.org/3/library/shutil.html#shutil.move
                shutil.move(str(new_file), str(destination_path))

            # Clean build base lib folder for next install
            shutil.rmtree(lib_dir_path, ignore_errors=True)


def copy_extra_files(extra_file_strings: Iterable[str]) -> None:
    """Copy extra files to their destinations."""
    if not extra_file_strings:
        return

    for extra_file_string in extra_file_strings:
        # Convert 'source > destination' strings to Paths.
        input_output = re.split(r' *> *', extra_file_string)
        source_file = Path(input_output[0])
        dest_file = Path(input_output[1])

        if not source_file.exists():
            raise FileNotFoundError(f'extra_file "{source_file}" not found.\n'
                                    f'  Defined by: "{extra_file_string}"')

        # Copy files and make parent directories.
        dest_file.parent.mkdir(parents=True, exist_ok=True)
        # Raise an error if the destination file already exists.
        if dest_file.exists():
            raise FileExistsError(
                f'Copying "{source_file}" would overwrite "{dest_file}"')

        shutil.copy(source_file, dest_file)


def main():
    args = _parse_args()

    py_packages = load_packages(args.input_list_files)

    build_python_tree(python_packages=py_packages,
                      tree_destination_dir=args.tree_destination_dir,
                      include_tests=args.include_tests)
    copy_extra_files(args.extra_files)

    if args.setupcfg_common_file:
        config = load_common_config(
            common_config=args.setupcfg_common_file,
            append_git_sha=args.setupcfg_version_append_git_sha,
            append_date=args.setupcfg_version_append_date)

        update_config_with_packages(config=config, python_packages=py_packages)

        write_config(common_config=args.setupcfg_common_file,
                     final_config=config,
                     tree_destination_dir=args.tree_destination_dir)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Creates a Git hook that calls a script with certain arguments."""

import argparse
import logging
import os
from pathlib import Path
import re
import shlex
import subprocess
from typing import Sequence, Union

_LOG: logging.Logger = logging.getLogger(__name__)


def git_repo_root(path: Union[Path, str]) -> Path:
    return Path(
        subprocess.run(['git', '-C', path, 'rev-parse', '--show-toplevel'],
                       check=True,
                       stdout=subprocess.PIPE).stdout.strip().decode())


def install_hook(script,
                 hook: str,
                 args: Sequence[str] = (),
                 repository: Union[Path, str] = '.') -> None:
    """Installs a simple Git hook that calls a script with arguments."""
    root = git_repo_root(repository).resolve()
    script = os.path.relpath(script, root)

    if root.joinpath('.git').is_dir():
        hook_path = root.joinpath('.git', 'hooks', hook)
    else:  # This repo is probably a submodule with a .git file instead
        match = re.match('^gitdir: (.*)$', root.joinpath('.git').read_text())
        if not match:
            raise ValueError('Unexpected format for .git file')

        hook_path = root.joinpath(match.group(1), 'hooks', hook).resolve()

    hook_path.parent.mkdir(exist_ok=True)

    command = ' '.join(shlex.quote(arg) for arg in (script, *args))

    with hook_path.open('w') as file:
        line = lambda *args: print(*args, file=file)

        line('#!/bin/sh')
        line(f'# {hook} hook generated by {__file__}')
        line()
        line('# Unset Git environment variables, which are set when this is ')
        line('# run as a Git hook. These environment variables cause issues ')
        line('# when trying to run Git commands on other repos from a ')
        line('# submodule hook.')
        line('unset $(git rev-parse --local-env-vars)')
        line()
        line(command)

    hook_path.chmod(0o755)
    logging.info('Created %s hook for %s at %s', hook, script, hook_path)


def argument_parser(parser=None) -> argparse.ArgumentParser:
    if parser is None:
        parser = argparse.ArgumentParser(description=__doc__)

    def path(arg: str) -> Path:
        if not os.path.exists(arg):
            raise argparse.ArgumentTypeError(f'"{arg}" is not a valid path')

        return Path(arg)

    parser.add_argument(
        '-r',
        '--repository',
        default='.',
        type=path,
        help='Path to the repository in which to install the hook')
    parser.add_argument('--hook',
                        required=True,
                        help='Which type of Git hook to create')
    parser.add_argument('-s',
                        '--script',
                        required=True,
                        type=path,
                        help='Path to the script to execute in the hook')
    parser.add_argument('args',
                        nargs='*',
                        help='Arguments to provide to the commit hook')

    return parser


if __name__ == '__main__':
    logging.basicConfig(format='%(message)s', level=logging.INFO)
    install_hook(**vars(argument_parser().parse_args()))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities for creating an interactive console."""

from collections import defaultdict
import functools
from itertools import chain
import inspect
import textwrap
import types
from typing import Any, Collection, Dict, Iterable, Mapping, NamedTuple

import pw_status
from pw_protobuf_compiler import python_protos

import pw_rpc
from pw_rpc.descriptors import Method
from pw_rpc.console_tools import functions

_INDENT = '    '


class CommandHelper:
    """Used to implement a help command in an RPC console."""
    @classmethod
    def from_methods(cls,
                     methods: Iterable[Method],
                     variables: Mapping[str, object],
                     header: str,
                     footer: str = '') -> 'CommandHelper':
        return cls({m.full_name: m
                    for m in methods}, variables, header, footer)

    def __init__(self,
                 methods: Mapping[str, object],
                 variables: Mapping[str, object],
                 header: str,
                 footer: str = ''):
        self._methods = methods
        self._variables = variables
        self.header = header
        self.footer = footer

    def help(self, item: object = None) -> str:
        """Returns a help string with a command or all commands listed."""

        if item is None:
            all_vars = '\n'.join(sorted(self._variables_without_methods()))
            all_rpcs = '\n'.join(self._methods)
            return (f'{self.header}\n\n'
                    f'All variables:\n\n{textwrap.indent(all_vars, _INDENT)}'
                    '\n\n'
                    f'All commands:\n\n{textwrap.indent(all_rpcs, _INDENT)}'
                    f'\n\n{self.footer}'.strip())

        # If item is a string, find commands matching that.
        if isinstance(item, str):
            matches = {n: m for n, m in self._methods.items() if item in n}
            if not matches:
                return f'No matches found for {item!r}'

            if len(matches) == 1:
                name, method = next(iter(matches.items()))
                return f'{name}\n\n{inspect.getdoc(method)}'

            return f'Multiple matches for {item!r}:\n\n' + textwrap.indent(
                '\n'.join(matches), _INDENT)

        return inspect.getdoc(item) or f'No documentation for {item!r}.'

    def _variables_without_methods(self) -> Mapping[str, object]:
        packages = frozenset(
            n.split('.', 1)[0] for n in self._methods if '.' in n)

        return {
            name: var
            for name, var in self._variables.items() if name not in packages
        }

    def __call__(self, item: object = None) -> None:
        """Prints the help string."""
        print(self.help(item))

    def __repr__(self) -> str:
        """Returns the help, so foo and foo() are equivalent in a console."""
        return self.help()


class ClientInfo(NamedTuple):
    """Information about an RPC client as it appears in the console."""
    # The name to use in the console to refer to this client.
    name: str

    # An object to use in the console for the client. May be a pw_rpc.Client.
    client: object

    # The pw_rpc.Client; may be the same object as client.
    rpc_client: pw_rpc.Client


def flattened_rpc_completions(
    client_info_list: Collection[ClientInfo], ) -> Dict[str, str]:
    """Create a flattened list of rpc commands for repl auto-completion.

    This gathers all rpc commands from a set of ClientInfo variables and
    produces a flattened list of valid rpc commands to run in an RPC
    console. This is useful for passing into
    prompt_toolkit.completion.WordCompleter.

    Args:
      client_info_list: List of ClientInfo variables

    Returns:
      Dict of flattened rpc commands as keys, and 'RPC' as values.
      For example: ::

        {
            'device.rpcs.pw.rpc.EchoService.Echo': 'RPC,
            'device.rpcs.pw.rpc.BatteryService.GetBatteryStatus': 'RPC',
        }
    """
    rpc_list = list(
        chain.from_iterable([
            '{}.rpcs.{}'.format(c.name, a.full_name)
            for a in c.rpc_client.methods()
        ] for c in client_info_list))

    # Dict should contain completion text as keys and descriptions as values.
    custom_word_completions = {
        flattened_rpc_name: 'RPC'
        for flattened_rpc_name in rpc_list
    }
    return custom_word_completions


class Context:
    """The Context class is used to set up an interactive RPC console.

    The Context manages a set of variables that make it easy to access RPCs and
    protobufs in a REPL.

    As an example, this class can be used to set up a console with IPython:

    .. code-block:: python

      context = console_tools.Context(
          clients, default_client, protos, help_header=WELCOME_MESSAGE)
      IPython.terminal.embed.InteractiveShellEmbed().mainloop(
          module=types.SimpleNamespace(**context.variables()))
    """
    def __init__(self,
                 client_info: Collection[ClientInfo],
                 default_client: Any,
                 protos: python_protos.Library,
                 *,
                 help_header: str = '') -> None:
        """Creates an RPC console context.

        Protos and RPC services are accessible by their proto package and name.
        The target for these can be set with the set_target function.

        Args:
          client_info: ClientInfo objects that represent the clients this
              console uses to communicate with other devices
          default_client: default client object; must be one of the clients
          protos: protobufs to use for RPCs for all clients
          help_header: Message to display for the help command
        """
        assert client_info, 'At least one client must be provided!'

        self.client_info = client_info
        self.current_client = default_client
        self.protos = protos

        # Store objects with references to RPC services, sorted by package.
        self._services: Dict[str, types.SimpleNamespace] = defaultdict(
            types.SimpleNamespace)

        self._variables: Dict[str, object] = dict(
            Status=pw_status.Status,
            set_target=functions.help_as_repr(self.set_target),
            # The original built-in help function is available as 'python_help'.
            python_help=help,
        )

        # Make the RPC clients and protos available in the console.
        self._variables.update((c.name, c.client) for c in self.client_info)

        # Make the proto package hierarchy directly available in the console.
        for package in self.protos.packages:
            self._variables[package._package] = package  # pylint: disable=protected-access

        # Monkey patch the message types to use an improved repr function.
        for message_type in self.protos.messages():
            message_type.__repr__ = python_protos.proto_repr

        # Set up the 'help' command.
        all_methods = chain.from_iterable(c.rpc_client.methods()
                                          for c in self.client_info)
        self._helper = CommandHelper.from_methods(
            all_methods, self._variables, help_header,
            'Type a command and hit Enter to see detailed help information.')

        self._variables['help'] = self._helper

        # Call set_target to set up for the default target.
        self.set_target(self.current_client)

    def flattened_rpc_completions(self):
        """Create a flattened list of rpc commands for repl auto-completion."""
        return flattened_rpc_completions(self.client_info)

    def variables(self) -> Dict[str, Any]:
        """Returns a mapping of names to variables for use in an RPC console."""
        return self._variables

    def set_target(self,
                   selected_client: object,
                   channel_id: int = None) -> None:
        """Sets the default target for commands."""
        # Make sure the variable is one of the client variables.
        name = ''
        rpc_client: Any = None

        for name, client, rpc_client in self.client_info:
            if selected_client is client:
                print('CURRENT RPC TARGET:', name)
                break
        else:
            raise ValueError('Supported targets :' +
                             ', '.join(c.name for c in self.client_info))

        # Update the RPC services to use the newly selected target.
        for service_client in rpc_client.channel(channel_id).rpcs:
            # Patch all method protos to use the improved __repr__ function too.
            for method in (m.method for m in service_client):
                method.request_type.__repr__ = python_protos.proto_repr
                method.response_type.__repr__ = python_protos.proto_repr

            service = service_client._service  # pylint: disable=protected-access
            setattr(self._services[service.package], service.name,
                    service_client)

        # Add the RPC methods to their proto packages.
        for package_name, rpcs in self._services.items():
            self.protos.packages[package_name]._add_item(rpcs)  # pylint: disable=protected-access

        self.current_client = selected_client


def _create_command_alias(command: Any, name: str, message: str) -> object:
    """Wraps __call__, __getattr__, and __repr__ to print a message."""
    @functools.wraps(command.__call__)
    def print_message_and_call(_, *args, **kwargs):
        print(message)
        return command(*args, **kwargs)

    def getattr_and_print_message(_, name: str) -> Any:
        attr = getattr(command, name)
        print(message)
        return attr

    return type(
        name, (),
        dict(__call__=print_message_and_call,
             __getattr__=getattr_and_print_message,
             __repr__=lambda _: message))()


def _access_in_dict_or_namespace(item, name: str, create_if_missing: bool):
    """Gets name as either a key or attribute on item."""
    try:
        return item[name]
    except KeyError:
        if create_if_missing:
            try:
                item[name] = types.SimpleNamespace()
                return item[name]
            except TypeError:
                pass
    except TypeError:
        pass

    if create_if_missing and not hasattr(item, name):
        setattr(item, name, types.SimpleNamespace())

    return getattr(item, name)


def _access_names(item, names: Iterable[str], create_if_missing: bool):
    for name in names:
        item = _access_in_dict_or_namespace(item, name, create_if_missing)

    return item


def alias_deprecated_command(variables: Any, old_name: str,
                             new_name: str) -> None:
    """Adds an alias for an old command that redirects to the new command.

    The deprecated command prints a message then invokes the new command.
    """
    # Get the new command.
    item = _access_names(variables,
                         new_name.split('.'),
                         create_if_missing=False)

    # Create a wrapper to the new comamnd with the old name.
    wrapper = _create_command_alias(
        item, old_name,
        f'WARNING: {old_name} is DEPRECATED; use {new_name} instead')

    # Add the wrapper to the variables with the old command's name.
    name_parts = old_name.split('.')
    item = _access_names(variables, name_parts[:-1], create_if_missing=True)
    setattr(item, name_parts[-1], wrapper)



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor = ...

# To simplify identifying the origin of a packet, client-to-server packets
# use even numbers and server-to-client packets use odd numbers.
class PacketType(_PacketType, metaclass=_PacketTypeEnumTypeWrapper):
    pass
class _PacketType:
    V = typing.NewType('V', builtins.int)
class _PacketTypeEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[_PacketType.V], builtins.type):
    DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor = ...
    # Client-to-server packets

    # The client invokes an RPC. Always the first packet.
    REQUEST = PacketType.V(0)
    # A message in a client stream. Always sent after a REQUEST and before a
    # CLIENT_STREAM_END.
    CLIENT_STREAM = PacketType.V(2)
    # The client received a packet for an RPC it did not request.
    CLIENT_ERROR = PacketType.V(4)
    # Deprecated, do not use. Send a CLIENT_ERROR with status CANCELLED instead.
    # TODO(pwbug/512): Remove this packet type.
    DEPRECATED_CANCEL = PacketType.V(6)
    # A client stream has completed.
    CLIENT_STREAM_END = PacketType.V(8)
    # Server-to-client packets

    # The RPC has finished.
    RESPONSE = PacketType.V(1)
    # Deprecated, do not use. Formerly was used as the last packet in a server
    # stream.
    # TODO(pwbug/512): Remove this packet type.
    DEPRECATED_SERVER_STREAM_END = PacketType.V(3)
    # The server was unable to process a request.
    SERVER_ERROR = PacketType.V(5)
    # A message in a server stream.
    SERVER_STREAM = PacketType.V(7)

# Client-to-server packets

# The client invokes an RPC. Always the first packet.
REQUEST = PacketType.V(0)
# A message in a client stream. Always sent after a REQUEST and before a
# CLIENT_STREAM_END.
CLIENT_STREAM = PacketType.V(2)
# The client received a packet for an RPC it did not request.
CLIENT_ERROR = PacketType.V(4)
# Deprecated, do not use. Send a CLIENT_ERROR with status CANCELLED instead.
# TODO(pwbug/512): Remove this packet type.
DEPRECATED_CANCEL = PacketType.V(6)
# A client stream has completed.
CLIENT_STREAM_END = PacketType.V(8)
# Server-to-client packets

# The RPC has finished.
RESPONSE = PacketType.V(1)
# Deprecated, do not use. Formerly was used as the last packet in a server
# stream.
# TODO(pwbug/512): Remove this packet type.
DEPRECATED_SERVER_STREAM_END = PacketType.V(3)
# The server was unable to process a request.
SERVER_ERROR = PacketType.V(5)
# A message in a server stream.
SERVER_STREAM = PacketType.V(7)
global___PacketType = PacketType


class RpcPacket(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor = ...
    TYPE_FIELD_NUMBER: builtins.int
    CHANNEL_ID_FIELD_NUMBER: builtins.int
    SERVICE_ID_FIELD_NUMBER: builtins.int
    METHOD_ID_FIELD_NUMBER: builtins.int
    PAYLOAD_FIELD_NUMBER: builtins.int
    STATUS_FIELD_NUMBER: builtins.int
    CALL_ID_FIELD_NUMBER: builtins.int
    # The type of packet. Determines which other fields are used.
    type: global___PacketType.V = ...
    # Channel through which the packet is sent.
    channel_id: builtins.int = ...
    # Hash of the fully-qualified name of the service with which this packet is
    # associated. For RPC packets, this is the service that processes the packet.
    service_id: builtins.int = ...
    # Hash of the name of the method which should process this packet.
    method_id: builtins.int = ...
    # The packet's payload, which is an encoded protobuf.
    payload: builtins.bytes = ...
    # Status code for the RPC response or error.
    status: builtins.int = ...
    # Unique identifier for the call that initiated this RPC. Optionally set by
    # the client in the initial request and sent in all subsequent client
    # packets; echoed by the server.
    call_id: builtins.int = ...
    def __init__(self,
        *,
        type : global___PacketType.V = ...,
        channel_id : builtins.int = ...,
        service_id : builtins.int = ...,
        method_id : builtins.int = ...,
        payload : builtins.bytes = ...,
        status : builtins.int = ...,
        call_id : builtins.int = ...,
        ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal[u"call_id",b"call_id",u"channel_id",b"channel_id",u"method_id",b"method_id",u"payload",b"payload",u"service_id",b"service_id",u"status",b"status",u"type",b"type"]) -> None: ...
global___RpcPacket = RpcPacket



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Decoder class for decoding bytes using HDLC protocol"""

import enum
import logging
from typing import Iterator, Optional
import zlib

from pw_hdlc import protocol

_LOG = logging.getLogger('pw_hdlc')

NO_ADDRESS = -1
_MIN_FRAME_SIZE = 6  # 1 B address + 1 B control + 4 B CRC-32


class FrameStatus(enum.Enum):
    """Indicates that an error occurred."""
    OK = 'OK'
    FCS_MISMATCH = 'frame check sequence failure'
    FRAMING_ERROR = 'invalid flag or escape characters'
    BAD_ADDRESS = 'address field too long'


class Frame:
    """Represents an HDLC frame."""
    def __init__(self,
                 raw_encoded: bytes,
                 raw_decoded: bytes,
                 status: FrameStatus = FrameStatus.OK):
        """Parses fields from an HDLC frame.

        Arguments:
            raw_encoded: The complete HDLC-encoded frame, excluding HDLC flag
                characters.
            raw_decoded: The complete decoded frame (address, control,
                information, FCS).
            status: Whether parsing the frame succeeded.
        """
        self.raw_encoded = raw_encoded
        self.raw_decoded = raw_decoded
        self.status = status

        self.address: int = NO_ADDRESS
        self.control: bytes = b''
        self.data: bytes = b''

        if status == FrameStatus.OK:
            address, address_length = protocol.decode_address(raw_decoded)
            if address_length == 0:
                self.status = FrameStatus.BAD_ADDRESS
                return

            self.address = address
            self.control = raw_decoded[address_length:address_length + 1]
            self.data = raw_decoded[address_length + 1:-4]

    def ok(self) -> bool:
        """True if this represents a valid frame.

        If false, then parsing failed. The status is set to indicate what type
        of error occurred, and the data field contains all bytes parsed from the
        frame (including bytes parsed as address or control bytes).
        """
        return self.status is FrameStatus.OK

    def __repr__(self) -> str:
        if self.ok():
            body = (f'address={self.address}, control={self.control!r}, '
                    f'data={self.data!r}')
        else:
            body = str(self.status)

        return f'{type(self).__name__}({body})'


class _State(enum.Enum):
    INTERFRAME = 0
    FRAME = 1
    FRAME_ESCAPE = 2


def _check_frame(frame_data: bytes) -> FrameStatus:
    if len(frame_data) < _MIN_FRAME_SIZE:
        return FrameStatus.FRAMING_ERROR

    frame_crc = int.from_bytes(frame_data[-4:], 'little')
    if zlib.crc32(frame_data[:-4]) != frame_crc:
        return FrameStatus.FCS_MISMATCH

    return FrameStatus.OK


class FrameDecoder:
    """Decodes one or more HDLC frames from a stream of data."""
    def __init__(self):
        self._decoded_data = bytearray()
        self._raw_data = bytearray()
        self._state = _State.INTERFRAME

    def process(self, data: bytes) -> Iterator[Frame]:
        """Decodes and yields HDLC frames, including corrupt frames.

        The ok() method on Frame indicates whether it is valid or represents a
        frame parsing error.

        Yields:
          Frames, which may be valid (frame.ok()) or corrupt (!frame.ok())
        """
        for byte in data:
            frame = self._process_byte(byte)
            if frame:
                yield frame

    def process_valid_frames(self, data: bytes) -> Iterator[Frame]:
        """Decodes and yields valid HDLC frames, logging any errors."""
        for frame in self.process(data):
            if frame.ok():
                yield frame
            else:
                _LOG.warning('Failed to decode frame: %s; discarded %d bytes',
                             frame.status.value, len(frame.raw_encoded))
                _LOG.debug('Discarded data: %s', frame.raw_encoded)

    def _finish_frame(self, status: FrameStatus) -> Frame:
        frame = Frame(bytes(self._raw_data), bytes(self._decoded_data), status)
        self._raw_data.clear()
        self._decoded_data.clear()
        return frame

    def _process_byte(self, byte: int) -> Optional[Frame]:
        frame: Optional[Frame] = None

        # Record every byte except the flag character.
        if byte != protocol.FLAG:
            self._raw_data.append(byte)

        if self._state is _State.INTERFRAME:
            if byte == protocol.FLAG:
                if self._raw_data:
                    frame = self._finish_frame(FrameStatus.FRAMING_ERROR)

                self._state = _State.FRAME
        elif self._state is _State.FRAME:
            if byte == protocol.FLAG:
                if self._raw_data:
                    frame = self._finish_frame(_check_frame(
                        self._decoded_data))

                self._state = _State.FRAME
            elif byte == protocol.ESCAPE:
                self._state = _State.FRAME_ESCAPE
            else:
                self._decoded_data.append(byte)
        elif self._state is _State.FRAME_ESCAPE:
            if byte == protocol.FLAG:
                frame = self._finish_frame(FrameStatus.FRAMING_ERROR)
                self._state = _State.FRAME
            elif byte in protocol.VALID_ESCAPED_BYTES:
                self._state = _State.FRAME
                self._decoded_data.append(protocol.escape(byte))
            else:
                self._state = _State.INTERFRAME
        else:
            raise AssertionError(f'Invalid decoder state: {self._state}')

        return frame
